#pragma once
#include "TProcuraMelhorativa.h"

///////////////////////////////////////////////////////////////////////////////
//	CEstacoes class
///////////////////////////////////////////////////////////////////////////////
//	Author: Jose Coelho
//	Last revision: 2024-05-09
//	Description:
//    Implementa o problema da colocao de estacoes
//    - A - minimizar o nmero de estaes
//    - B - minimizar o custo mdio da populao  estao mais prxima
//    - B tem de ser menor que 3.
//    Custo, com base na distncia (diagonal igual a horizontal/vertical):
//    - 0 a 1: custo 0
//    - 2 <= K <= 5: custo 2^(K-2), ou seja, de 2..5: 1, 2, 4, 8
//    - 6 ou superior: custo 10
//    Soluo: zonas das estaes colocadas
//    Abordagem construtiva:
//    - comea sem estaes, e cada ao coloca uma estao 
//    - melhor primeiro pode no ser muito apropriado 
//    - no  facil obter heursticas admissveis
//    Abordagem melhorativa:
//    - permutao de zonas (gerador aleatrio)
//    - coloca cada estao por ordem, at que B seja menor que 3
//    - remover da permutao as restantes zonas
//    Vizinhana:
//    - mover uma estao uma unidade para qualquer direo
//    Mutao:
//    - mover nmero arbitrrio de estaes (multiplas vizinhanas)
//    Cruzamento:
//    - processar estaes por ordem arbitrria, de ambas as solues
//    - parar assim que B seja inferior a 3
//    Correo aps alterao de posies:
//    - caso B fique maior que 3, adicionar nova estao
//    - testar a remoo de uma estao, se B ficar ainda menor que 3, remover
//    Otimizao:
//    - estaes tm de estar pelo menos a distncia 2, entre si
//    - heurstica: fixar uma distncia K, maior que 2, entre estaes
///////////////////////////////////////////////////////////////////////////////

#define MAX_LARGURA 100
#define MAX_ALTURA 100

#define _MAX(a,b) (a>b?a:b)
#define _MIN(a,b) (a<b?a:b)

class TZona {
public:
	int linha, coluna; // coordenadas de uma zona
	TZona(int linha=0, int coluna=0) {
		TZona::linha = linha;
		TZona::coluna = coluna;
	}
	~TZona() {}
	bool operator==(TZona& z) { return linha == z.linha && coluna == z.coluna; }
	bool operator<(TZona& z) { return linha < z.linha || linha == z.linha && coluna < z.coluna; }
	int Distancia(TZona z) { return _MAX(abs(linha - z.linha), abs(coluna - z.coluna));	}
};

class CEstacao :
	public TProcuraMelhorativa
{
private:
	int Distancia(TZona &zona); // retorna a distncia at  estao mais prxima
	int Custo(TZona &zona); // retorna o custo com base na distncia at  estao mais prxima

public:
	CEstacao(void);
	~CEstacao(void);

	// dados da instncia (estticos)
	static int largura, altura; // dimenses da zona
	static int populacao[MAX_ALTURA][MAX_LARGURA]; // populao (caregado uma s vez para a instncia)

	static void CarregarInstancia(int id);
	static int distMin; // distncia mnima entre estaes
	static int familias; // nmero total de famlias

	// soluo
	TVector<TZona> estacao;
	// decises: construir estacao num local

	// metodos redefinidos de TProcuraConstrutiva

	// Cria um objecto que e uma copia deste
	TProcuraConstrutiva* Duplicar(void);
	void Copiar(TProcuraConstrutiva* objecto);

	// Coloca o objecto no estado inicial da procura
	void SolucaoVazia(void);

	// Coloca em sucessores a lista de objectos sucessores (sao alocados neste metodo e tem de ser apagados)
	// O custo nao necessita de ser preenchido, caso seja sempre unitario
	void Sucessores(TVector<TProcuraConstrutiva*>&sucessores, TVector<int>&custo);
	// Retorna verdade caso o estado actual seja um estado objectivo
	bool SolucaoCompleta(void);
	// Escrever informacao de debug sobre o objecto currente
	// (utilizar variavel TProcuraConstrutiva::debug para seleccionar o detalhe pretendido)
	void Debug(void);
	int Heuristica(void);

	// abrodagem melhorativa
	void NovaSolucao(void);
	int Avaliar(void);
	void Vizinhanca(TVector<TProcuraMelhorativa*>& vizinhos);
	void Mutar(void);
	void Cruzamento(TProcuraMelhorativa* a, TProcuraMelhorativa* b);
	int Distancia(TProcuraMelhorativa* a);

};
